"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const p5_analysis_1 = require("p5-analysis");
const helpers_1 = require("../helpers");
const ts_extras_1 = require("../ts-extras");
const fs_1 = __importDefault(require("fs"));
const path_1 = __importDefault(require("path"));
const sketchTypes = {
    '^html$': 'html',
    '^(script|javascript|js)(-only)?$': 'script',
    '^folder$': 'folder'
};
async function convert(sketchPath, options) {
    if (!options.to) {
        (0, helpers_1.die)(`Missing required option: --to`);
    }
    const targetType = Object.entries(sketchTypes).find(([regex]) => options.to.match(regex))?.[1] ??
        (0, helpers_1.die)(`Invalid option --to ${options.to}`);
    if (!fs_1.default.existsSync(sketchPath) && fs_1.default.existsSync(sketchPath + '.html'))
        sketchPath += '.html';
    if (!fs_1.default.existsSync(sketchPath) && fs_1.default.existsSync(sketchPath + '.js'))
        sketchPath += '.js';
    // TODO: if it's a script file that belongs to an HTML index in the same directory, warn or rename the index instead
    const sketch = await p5_analysis_1.Sketch.fromFile(sketchPath);
    if (targetType === 'folder') {
        const targetDir = sketchPath.replace(/\.(html?|js)/i, '');
        fs_1.default.mkdirSync(targetDir, { recursive: true });
        sketch.files.forEach(file => {
            const targetName = file === sketch.htmlFile ? 'index.html' : file;
            fs_1.default.renameSync(path_1.default.join(sketch.dir, file), path_1.default.join(targetDir, targetName));
            console.log(`Moved ${file} into new directory ${targetDir}`);
        });
        return;
    }
    if (sketch.structureType === targetType) {
        console.log('Nothing to do');
        return;
    }
    try {
        await sketch.convert({ type: targetType });
    }
    catch (err) {
        (0, ts_extras_1.assertError)(err);
        (0, helpers_1.die)(err.message);
    }
}
exports.default = convert;
