"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const p5_analysis_1 = require("p5-analysis");
const helpers_1 = require("../helpers");
const ts_extras_1 = require("../ts-extras");
const fs_1 = __importDefault(require("fs"));
const path_1 = __importDefault(require("path"));
async function create(file, options) {
    let scriptFile;
    if (options.type && !options.type.match(/folder|file/)) {
        (0, helpers_1.die)('create: type must be "folder" or "file"');
    }
    if (options.type === 'folder' || !/\.(js|html?)$/i.test(file)) {
        try {
            fs_1.default.mkdirSync(file);
        }
        catch (err) {
            (0, ts_extras_1.assertError)(err);
            if (err.code !== 'EEXIST') {
                throw err;
            }
            if (!fs_1.default.statSync(file).isDirectory()) {
                (0, helpers_1.die)(`The ${file} folder already exists and is not a directory`);
            }
            // TODO: ignore .DS_Store and Thumbs.db
            if (fs_1.default.readdirSync(file).length && !options.force) {
                (0, helpers_1.die)(`The ${file} folder already exists and is not empty`);
            }
        }
        file = path_1.default.join(file, 'index.html');
    }
    else if (/\.html?$/i.test(file)) {
        scriptFile = path_1.default.basename(file).replace(/\.html?$/i, '.js');
    }
    const templateOptions = (0, helpers_1.stringToOptions)(options.options);
    const sketchOptions = { scriptFile, ...options };
    const sketch = p5_analysis_1.Sketch.create(file, sketchOptions);
    let files = [];
    try {
        files = await sketch.generate(options.force, templateOptions);
    }
    catch (err) {
        (0, ts_extras_1.assertError)(err);
        if (err.code === 'EEXIST') {
            (0, helpers_1.die)(`${file} already exists. Try again with --force.`);
        }
        console.error(Object.entries(err));
        throw err;
    }
    files.forEach(file => console.log(`Created ${file}`));
}
exports.default = create;
