"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const promises_1 = require("fs/promises");
const p5_analysis_1 = require("p5-analysis");
const posix_1 = __importDefault(require("path/posix"));
const Server_1 = require("../server/Server");
const helpers_1 = require("../helpers");
const typeMap = {
    jpeg: 'jpeg',
    jpg: 'jpeg',
    png: 'png'
};
async function screenshot(source, options) {
    if (await p5_analysis_1.Sketch.isSketchDir(source)) {
        // const sketch = await Sketch.fromDirectory(source);
        // source = sketch.mainFilePath;
    }
    else if (!(await p5_analysis_1.Sketch.isSketchFile(source))) {
        (0, helpers_1.die)(`${source} is not a sketch file`);
    }
    const output = options.output ||
        posix_1.default
            .basename(source.replace(/(.+)\/index\.html?/i, '$1'))
            .replace(/\.(js|html?)$/i, '') +
            (Number(options.frameCount || 1) > 1 ? '-%d.png' : '.png');
    const ext = output.split('.').pop() ?? 'png';
    const imageType = typeMap[ext.toLowerCase()];
    if (!imageType) {
        (0, helpers_1.die)('The output file extension must be .png or .jpeg');
    }
    const serverOptions = {
        root: source,
        screenshot: {
            onFrameData,
            imageType,
            ...parseScreenshotOptions(options)
        }
    };
    let remainingFrames = serverOptions.screenshot?.frameCount || 1;
    if (remainingFrames > 1 && !/%\d*d/.test(output)) {
        console.warn('Warning: For best results, include a %d in the output filename when capturing multiple frames');
    }
    const server = await Server_1.Server.start(serverOptions);
    (0, helpers_1.openInBrowser)(server.url, options.browser?.toLowerCase());
    async function onFrameData({ data, frameNumber }) {
        if (remainingFrames < 0)
            return;
        const fname = output.replace(/%\d*d/g, fmt => {
            let s = String(frameNumber);
            const m = fmt.match(/%(0)?(\d+)/);
            if (m) {
                const pad = m[1] || '0';
                const len = Number(m[2]);
                s = s.padStart(len, pad);
            }
            return s;
        });
        await (0, promises_1.mkdir)(posix_1.default.dirname(fname), { recursive: true });
        await (0, promises_1.writeFile)(fname, data);
        console.log(`Saved screenshot from ${source} to ${fname}`);
        if (--remainingFrames == 0) {
            // Give the client time to receive the request response, so that it knows
            // to close.
            setTimeout(() => process.exit(0), 100);
        }
    }
}
exports.default = screenshot;
function parseScreenshotOptions(options) {
    const skipFrames = Number(options.skipFrames || 0);
    let canvasDimensions = undefined;
    if (options.canvasSize) {
        const m = options.canvasSize.match(/^(\d+)(?:[x, ](\d+))?$/);
        if (!m) {
            (0, helpers_1.die)(`Invalid canvas size: ${options.canvasSize}`);
        }
        canvasDimensions = { width: Number(m[1]), height: Number(m[2] ?? m[1]) };
    }
    let pixelDensity = undefined;
    if (options.pixelDensity) {
        const m = options.pixelDensity.match(/^(\d+(?:\.\d*)?|\.\d+)(?:\/(\d+(?:\.\d*)?|\.\d+))?$/);
        if (!m) {
            (0, helpers_1.die)(`Invalid pixel density: ${options.pixelDensity}`);
        }
        pixelDensity = Number(m[1]) / Number(m[2] || 1);
    }
    const frameCount = Number(options.frameCount || 1);
    return {
        canvasDimensions,
        frameCount,
        pixelDensity,
        skipFrames
    };
}
