/** Print the message to standard output; then exit with status code 1.
 */
export declare function die(message: string): never;
/** Escape HTML special characters in a string.
 */
export declare function escapeHTML(str: string): string;
/**
 *	Open a URL in the browser.
 *
 * This is a wrapper for `open()` from the 'open' package. It has the same API,
 *	 except that it also accepts 'safari' as an app name.
 */
export declare function openInBrowser(url: string, browser?: string): Promise<import('child_process').ChildProcess>;
/**
 * 'a/b/c' => [{name: 'Home', path: '/'}, {name: 'a', path: '/a'}, {name: 'b', path: '/a/b'}, {name: 'c', path: '/a/b/c'}]
 */
export declare function pathComponentsForBreadcrumbs(relDirPath: string): {
    name: string;
    path: string;
}[];
/** Tests whether filepath is inside the directory `dir`. */
export declare function pathIsInDirectory(filepath: string, dir: string): boolean;
/** Returns true iff pathname ends in a markdown file suffix. */
export declare function pathIsMarkdown(filepath: string): boolean;
/**
 * Examples:
 * 'a,b,c' => {a: true, b: true, c: true}
 * 'a,no-b,c' => {a: true, b: false, c: true}
 */
export declare function stringToOptions(str: string | null): {
    [k: string]: boolean;
};
/** Insert a <script> element in an HTML document's head.
 *
 * If the source argument is a string, it becomes the value of the element's
 * `src` attribute.
 *
 * If it is an object with a key `script` it becomes the text content of the
 * element.
 *
 * Otherwise it is an Object; the script tag defines its keys as global
 * variables, that are initialized to the corresponding values.
 */
export declare function addScriptToHtmlHead(html: string, source: string | Record<string, unknown>): string;
