/// <reference types="node" />
import express = require('express');
export * as cacache from 'cacache';
export declare const proxyPrefix = "/__p5_proxy_cache";
export declare const cachePath: string;
/** The express.Request properties that cdnProxyRouter depends on. */
interface RequestI {
    headers: typeof express.request.headers;
    path: typeof express.request.path;
    query: typeof express.request.query;
}
/** The express.Response properties that cdnProxyRouter depends on. */
interface ResponseI extends NodeJS.WritableStream {
    setHeader(key: string, value: string | number | readonly string[]): void;
    send(chunk: string | Buffer): void;
    status(code: number): void;
}
export declare function cdnProxyRouter(req: RequestI, res: ResponseI): Promise<void>;
export declare function encodeProxyPath(originUrl: string, { includePrefix }?: {
    includePrefix?: boolean | undefined;
}): string;
export declare function decodeProxyPath(proxyPath: string, query?: RequestI['query']): string;
export declare type CacheWarmStats = {
    total: number;
    failures: number;
    hits: number;
    misses: number;
};
export declare type CacheWarmMessage = {
    type: 'initial';
    total: number;
} | {
    type: 'prefetch';
    url: string;
} | {
    type: 'error';
    url: string;
    status: number;
} | {
    type: 'progress';
    stats: CacheWarmStats;
};
/** Warm the cache, by requesting all the urls in the manifest, and the urls that they reference.
 *
 * Most of the complexity come from fetching the URLs in parallel.
 *
 * (Currently, only references in CSS files are prefetched.)
 */
export declare function warmCache({ force }: {
    force?: boolean;
}, callback?: (message: CacheWarmMessage) => void): Promise<CacheWarmStats>;
export declare function isCdnUrl(url: string): boolean;
/** Replace CDN URLs in script[src] and link[href] with proxy cache paths.
 *
 * @param html the HTML to process
 * @returns the processed HTML
 */
export declare function replaceUrlsInHtml(html: string): string;
