"use strict";
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    Object.defineProperty(o, k2, { enumerable: true, get: function() { return m[k]; } });
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.replaceUrlsInHtml = exports.isCdnUrl = exports.warmCache = exports.decodeProxyPath = exports.encodeProxyPath = exports.cdnProxyRouter = exports.cachePath = exports.proxyPrefix = exports.cacache = void 0;
const cacache = __importStar(require("cacache"));
const csstree = __importStar(require("css-tree"));
const css_tree_1 = require("css-tree");
const node_fetch_1 = __importDefault(require("node-fetch"));
const node_html_parser_1 = require("node-html-parser");
const p5_analysis_1 = require("p5-analysis");
const Library_1 = require("p5-analysis/dist/models/Library");
const stream_1 = __importStar(require("stream"));
const zlib_1 = __importDefault(require("zlib"));
const ts_extras_1 = require("../ts-extras");
const path = require("path");
const assert = require("assert");
exports.cacache = __importStar(require("cacache"));
// eslint-disable-next-line @typescript-eslint/no-var-requires
const debug = require('debug')('p5-server:cdnProxy');
exports.proxyPrefix = '/__p5_proxy_cache';
exports.cachePath = process.env.HOME + '/.cache/p5-server';
const HTTP_RESPONSE_HEADER_CACHE_STATUS = 'x-p5-server-cache-hit';
const uncacheableResponseHeaders = [
    'accept-ranges', 'age', 'connection', 'content-accept-ranges',
    'content-length', 'strict-transport-security', 'transfer-encoding', 'vary'
];
/** A list of CDNs that aren't listed in the p5-analysis Library model (because
 * they aren't specific to serving NPM packages).
 */
const cdnDomains = [
    'fonts.googleapis.com',
    'fonts.gstatic.com',
    // JSDelivr is a known npm package proxy, but the templates use a different path schema to request Highlight.js
    // distribution files.
    'cdn.jsdelivr.net',
];
/** URLs to warm the cache with, that can't be inferred from the libraries, in
 * addition to library loadPaths.
 *
 * It's okay if these have repeats. The cache warmer deduplicates URLs anyway.
 */
const cacheSeeds = [
    // TODO: use an API to retrieve this constant
    `https://cdn.jsdelivr.net/npm/p5@${Library_1.p5Version}/lib/p5.min.js`,
    // TODO: read the following from the template file. Or, add these to the package.
    // directory.pug
    'https://cdn.jsdelivr.net/npm/jquery@3.6/dist/jquery.min.js',
    'https://cdn.jsdelivr.net/npm/semantic-ui@2.4/dist/semantic.min.js',
    'https://cdn.jsdelivr.net/npm/semantic-ui@2.4/dist/semantic.min.css',
    'https://cdn.jsdelivr.net/gh/highlightjs/cdn-release@11.2.0/build/styles/default.min.css',
    // markdown.pug
    'https://cdn.jsdelivr.net/npm/semantic-ui@2.4/dist/semantic.min.css',
    'https://cdn.jsdelivr.net/gh/highlightjs/cdn-release@11.2.0/build/styles/default.min.css',
    // source-view.pug
    "https://cdn.jsdelivr.net/gh/highlightjs/cdn-release@11.2.0/build/styles/github-dark.min.css",
    "https://cdn.jsdelivr.net/gh/highlightjs/cdn-release@11.2.0/build/highlight.min.js",
];
// Note that express.Request implements RequestI, and express.Response
// implements ResponseI.
//
// This function uses the more general type to allow prefetch to call
// cdnProxyRouter.
async function cdnProxyRouter(req, res) {
    const originUrl = decodeProxyPath(req.path, req.query);
    // An earlier version used a cryptographic digest of the stringified JSON;
    // however, the 'crypto' module is not present in VSCode.
    const cacheKey = JSON.stringify({
        url: originUrl,
        // Formally, the cache key should include the headers in the Vary response
        // header. In practice, this header only has at most the following keys;
        // and, it is harmless to vary on them even when they aren't specified.
        //
        // Do NOT cache on User-Agent. It is not necessary for the supported CDNs,
        // and it would bust the cache between different browsers, which is
        // undesireable for offline development.
        accept: req.headers['accept'],
        acceptCh: req.headers['accept-ch'],
        acceptEncoding: req.headers['accept-encoding'],
        acceptLanguage: req.headers['accept-language'],
    });
    const cacheObject = await cacache.get.info(exports.cachePath, cacheKey);
    res.setHeader('x-p5-server-origin-url', originUrl);
    if (cacheObject && !req.query.reload) {
        debug('cache hit', originUrl);
        res.setHeader(HTTP_RESPONSE_HEADER_CACHE_STATUS, 'HIT');
        const { headers } = cacheObject.metadata;
        for (const key of Object.keys(headers)) {
            let value = headers[key];
            switch (key) {
                case 'location':
                    if (isCdnUrl(value)) {
                        value = encodeProxyPath(value);
                    }
                    break;
                case 'cache-control':
                    continue;
            }
            const headerMap = { server: 'origin-server' };
            res.setHeader(headerMap[key] ?? key, value);
        }
        {
            const age = Math.max(0, (+new Date()) - cacheObject.time);
            res.setHeader('age', Math.floor(age / 1000));
        }
        {
            let cacheControl = headers['cache-control']?.match(/(?:^|\b)(public|private)\b/)?.[1] ?? 'public';
            const maxAge = headers['cache-control']?.match(/(?:^|\b)max-age=(\d+)/)?.[1];
            if (maxAge) {
                cacheControl += ', max-age=' + maxAge;
            }
            cacheControl += `, stale-while-revalidate=${maxAge || 86400}`;
            res.setHeader('cache-control', cacheControl);
        }
        res.status(cacheObject.metadata.status);
        let rstream = cacache.get.stream(exports.cachePath, cacheKey);
        rstream = makeProxyReplacemenStream(rstream, headers['content-type'], headers['content-encoding'], originUrl);
        rstream.pipe(res);
        // check for cache expiration
        // This maxAge differs from the one above in that it prefers s-maxage over max-age if the former exists.
        const maxAge = (headers['cache-control']?.match(/(?:%|\b)s-maxage=(\d+)/) ||
            headers['cache-control']?.match(/(?:%|\b)max-age=(\d+)/))?.[1] ?? 'Infinity';
        const expires = new Date(cacheObject.time + Number(maxAge) * 1000);
        const expired = expires < new Date();
        if (expired) {
            debug('cache expired', originUrl);
            // The response is complete. Replace the original response instance with one that simply ignores writes.
            // This reduces the number of paths below.
            res = new class extends stream_1.default.Writable {
                /* eslint-disable @typescript-eslint/no-empty-function */
                setHeader() { }
                send() { }
                status() { }
                _write() { }
            };
        }
        else {
            await new Promise(resolve => res.on('finish', resolve));
            return;
        }
    }
    debug('proxy request for', originUrl);
    // filter the headers, and combine string[] values back into strings
    const headerAcceptList = ['accept', 'user-agent', 'accept-language', 'accept-encoding'];
    const reqHeaders = Object.fromEntries(Object.entries(req.headers)
        .filter(([key]) => headerAcceptList.includes(key))
        .filter(([_key, value]) => (0, ts_extras_1.isDefined)(value))
        .map(([key, value]) => [key, Array.isArray(value) ? value.join(',') : value]));
    const originResponse = await (0, node_fetch_1.default)(originUrl, {
        compress: false,
        headers: reqHeaders,
        redirect: 'manual', // don't follow redirects; cache the redirect directive instead
    });
    // Relay the origin status, and add a cache header
    res.status(originResponse.status);
    res.setHeader(HTTP_RESPONSE_HEADER_CACHE_STATUS, 'MISS');
    // Copy headers from the origin response to the output response.
    // Modify Location headers to proxy them, in the case of a redirect.
    originResponse.headers.forEach((value, key) => {
        if (key === 'location' && isCdnUrl(value)) {
            value = encodeProxyPath(value);
        }
        res.setHeader(key, value);
    });
    // This test excludes 300 Multiple Choice, since that status code is rarely
    // used in practice, and would require rewriting the links in the HTML
    // response.
    const redirected = 300 < originResponse.status && originResponse.status < 400 && originResponse.headers.has('location');
    if (!originResponse.ok && !redirected) {
        // don't cache responses other than 200's and redirects
        debug(`Failed ${originResponse.ok} | ${originResponse.status} | ${originResponse.statusText}`);
        res.send(originResponse.statusText);
        return;
    }
    // expressjs.Response.headers serializes to {}. Copy it to an Object that can
    // be serialized to JSON.
    const responseHeaders = Object.fromEntries(Array.from(originResponse.headers.entries())
        .filter(([key]) => !uncacheableResponseHeaders.includes(key)));
    const cacheWriteStream = cacache.put.stream(exports.cachePath, cacheKey, {
        metadata: {
            originUrl,
            headers: responseHeaders,
            status: originResponse.status
        }
    });
    // pipe the origin response body to both the client response and the cache
    // write stream. Collect the length of the response for logging.
    const streamLengthCounter = new class extends stream_1.default.Writable {
        constructor() {
            super(...arguments);
            this.length = 0;
        }
        _write(chunk, _encoding, callback) {
            if (typeof chunk === 'string' || chunk instanceof Buffer || chunk instanceof Uint8Array) {
                this.length += chunk.length;
            }
            callback();
        }
    };
    originResponse.body.pipe(multiplexStreamWriter([cacheWriteStream, streamLengthCounter]));
    makeProxyReplacemenStream(originResponse.body, responseHeaders['content-type'], responseHeaders['content-encoding'], originUrl).pipe(res);
    await new Promise(resolve => streamLengthCounter.on('finish', resolve));
    debug('wrote', streamLengthCounter.length, 'bytes to cache for', originUrl);
}
exports.cdnProxyRouter = cdnProxyRouter;
function makeProxyReplacemenStream(stream, contentType, contentEncoding, base) {
    if (contentType.startsWith('text/css')) {
        return makeCssRewriterStream(stream, base, contentEncoding);
    }
    return stream;
}
//#region proxy paths
// exported for unit testing
function encodeProxyPath(originUrl, { includePrefix = true } = {}) {
    if (!/^https?:/i.test(originUrl))
        return originUrl;
    let proxyPath = originUrl;
    if (/\?/.test(originUrl)) {
        // package the entire query string into a single query parameter, so that other query parameters can be added to the
        // URL without breaking the cache
        const u = new URL(originUrl);
        u.search = `?search=${encodeURIComponent(u.search.substr(1))}`;
        proxyPath = u.toString();
    }
    // The following transformation improves the readability of the developer console's source list.
    proxyPath = proxyPath
        .replace(/^https:\/\//i, '')
        .replace(/^http:\/\//i, 'http/');
    return includePrefix ? `${exports.proxyPrefix}/${proxyPath}` : proxyPath;
}
exports.encodeProxyPath = encodeProxyPath;
// exported for unit testing
function decodeProxyPath(proxyPath, query = {}) {
    let originUrl = proxyPath
        .replace(exports.proxyPrefix, '')
        .replace(/^\//, '')
        .replace(/^http\//, 'http://');
    if (!/^https?:/i.test(originUrl))
        originUrl = `https://${originUrl}`;
    if (originUrl.includes('?')) {
        const [pʹ, queryString, hash] = originUrl.match(/(.+)\?(.+)(#.*)?/).slice(1);
        originUrl = pʹ + (hash || '');
        new URLSearchParams(queryString).forEach((value, key) => {
            query[key] = value;
        });
    }
    if (query.search) {
        originUrl += `?${decodeURIComponent(query.search)}`;
    }
    return originUrl;
}
exports.decodeProxyPath = decodeProxyPath;
function isProxyPath(url) {
    return url.startsWith(exports.proxyPrefix);
}
//#endregion
//#region cache warmup
/** Verify that url is in the cache. Request it if it is not. Uses
 * cdnProxyRouter to minimize different code paths that need to be tested.
 * Returns a Response-like structure, that warmCache can use to follow
 * referenced URLs.
 */
async function prefetch(url, { accept = '*/*', force = false }) {
    const reqHeaders = {
        accept,
        // TODO: use a different user-agent for prefetching?
        'user-agent': 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/15.1 Safari/605.1.15',
        'accept-language': 'en-US,en;q=0.9',
        'accept-encoding': 'gzip, deflate',
    };
    const req = {
        headers: reqHeaders,
        path: encodeProxyPath(url, { includePrefix: false }),
        query: force ? { reload: 'true' } : {}
    };
    let status;
    /* eslint-disable @typescript-eslint/no-empty-function */
    const res = new class extends stream_1.default.Writable {
        constructor() {
            super(...arguments);
            this.headers = {};
            this.chunks = new Array();
        }
        setHeader(key, value) {
            this.headers[key] = value;
        }
        status(statusCode) { status = statusCode; }
        send(chunk) { this.chunks.push(typeof chunk === 'string' ? Buffer.from(chunk) : chunk); }
        _write(chunk, _encoding, callback) {
            assert.ok(chunk instanceof Buffer);
            this.chunks.push(chunk);
            callback();
        }
    };
    /* eslint-enable @typescript-eslint/no-empty-function */
    debug(`warm cache for ${url}`);
    await cdnProxyRouter(req, res);
    const redirected = 300 < status && status < 400 && res.headers.location.startsWith(exports.proxyPrefix + '/');
    if (redirected) {
        const location = decodeURIComponent(res.headers.location.substring(exports.proxyPrefix.length + 1));
        debug(`following redirect from ${url} -> ${location}`);
        return prefetch(location, { accept, force });
    }
    return {
        data: Buffer.concat(res.chunks),
        headers: res.headers,
        ok: status < 400,
        status: status,
    };
}
/** Warm the cache, by requesting all the urls in the manifest, and the urls that they reference.
 *
 * Most of the complexity come from fetching the URLs in parallel.
 *
 * (Currently, only references in CSS files are prefetched.)
 */
async function warmCache({ force }, callback) {
    const concurrency = 20; // max number of requests to make at once
    const stats = { total: 0, failures: 0, hits: 0, misses: 0 };
    // use Set to dedupe
    const urls = dedupe([...cacheSeeds, ...getLibraryImportPaths()]).sort();
    callback?.({ type: 'initial', total: urls.length });
    const seen = new Set();
    const promises = [];
    // `while` instead of `for`, because visit() can add to the array.
    while (urls.length > 0 || promises.length > 0) {
        const url = urls.shift();
        if (url) {
            if (seen.has(url))
                continue;
            seen.add(url);
            callback?.({ type: 'prefetch', url });
            await visit(url);
        }
        else {
            // One of the pending promises could add more urls to the queue, so wait
            // for the next one inside the loop instead of awaiting Promise.all()
            // at the end.
            await Promise.race(promises);
        }
    }
    return stats;
    // This function returns immediately once it adds a fetch promise to the
    // array. (It does not wait for the fetch to initiate.) If there are already
    // `concurrency` promises pending, it waits for one to resolve before adding
    // the new promise and returning.
    async function visit(url) {
        if (promises.length >= concurrency) {
            // debug('waiting for one of', promises.length, 'prefetches to settle');
            /* eslint-disable-next-line @typescript-eslint/no-empty-function */
            await Promise.race(promises);
        }
        const accept = {
            '.css': 'text/css',
            '.html': 'text/html',
        }[path.extname(url)] || '*/*';
        const p = prefetch(url, { accept, force })
            .then(({ status, ok, headers, data }) => {
            if (ok) {
                const hit = headers[HTTP_RESPONSE_HEADER_CACHE_STATUS] === 'HIT';
                if (hit)
                    stats.hits++;
                else
                    stats.misses++;
                // add this document's URLs to the list of URLs to prefetch
                if (headers['content-type'].startsWith('text/css') && data.length > 0) {
                    if (headers['content-encoding'] === 'gzip') {
                        data = zlib_1.default.gunzipSync(data);
                    }
                    const base = url;
                    cssForEachUrl(data.toString(), (value) => {
                        if (value.startsWith('data:'))
                            return;
                        // prefetch returns a document with the URLs replaced, so CDN URLs
                        // appear as proxy paths (or relative URLs), not as URLs with CDN
                        // hostnames.
                        value = removeHash(value);
                        if (isProxyPath(value)) {
                            const originUrl = decodeProxyPath(url);
                            urls.push(originUrl);
                        }
                        else if (isRelativeUrl(value)) {
                            const originUrl = urlResolve(base, value);
                            urls.push(originUrl);
                        }
                    });
                }
                callback?.({ type: 'progress', stats });
            }
            else {
                stats.failures++;
                callback?.({ type: 'error', url, status });
            }
        })
            .finally(() => {
            stats.total++;
            promises.splice(promises.indexOf(p), 1);
            // callback?.(stats);
        });
        promises.push(p);
    }
}
exports.warmCache = warmCache;
//#endregion
//#region CDN recognition
// exported for unit tests
function isCdnUrl(url) {
    if (!/^https?:/.test(url))
        return false;
    return p5_analysis_1.Cdn.all.some(cdn => cdn.matchesUrl(url))
        || getLibraryImportPaths().has(url)
        || cdnDomains.includes(new URL(url).hostname);
}
exports.isCdnUrl = isCdnUrl;
/** Cache for memoizing getLibraryImportPaths. */
let _libraryImportPaths;
function getLibraryImportPaths() {
    _libraryImportPaths ?? (_libraryImportPaths = new Set(p5_analysis_1.Library.all.map(lib => lib.importPath).filter(ts_extras_1.isDefined)));
    return _libraryImportPaths;
}
//#endregion
//#region rewrite documents
/** Replace CDN URLs in script[src] and link[href] with proxy cache paths.
 *
 * @param html the HTML to process
 * @returns the processed HTML
 */
function replaceUrlsInHtml(html) {
    const htmlRoot = (0, node_html_parser_1.parse)(html);
    let modified = false;
    // rewrite script[src]
    htmlRoot
        .querySelectorAll('script[src]')
        .filter(e => isCdnUrl(e.attributes.src))
        .forEach(e => {
        modified = true;
        e.setAttribute('src', encodeProxyPath(e.attributes.src));
    });
    // rewrite link[href]
    htmlRoot
        .querySelectorAll('link[rel=stylesheet][href]')
        .filter(e => isCdnUrl(e.attributes.href))
        .forEach(e => {
        modified = true;
        e.setAttribute('href', encodeProxyPath(e.attributes.href));
    });
    return modified ? htmlRoot.outerHTML : html;
}
exports.replaceUrlsInHtml = replaceUrlsInHtml;
/** Replace CDN URLs with proxy cache paths.
 *
 * @param html the HTML to process
 * @returns the processed HTML
 */
function replaceUrlsInCss(text) {
    const stylesheet = (0, css_tree_1.parse)(text);
    let modified = false;
    cssForEachUrl(stylesheet, value => {
        if (value.startsWith('data:'))
            return;
        if (isCdnUrl(value)) {
            const proxied = encodeProxyPath(value);
            modified = true;
            // debug(`rewriting ${value} to ${proxied}`);
            return proxied;
        }
    });
    return modified ? csstree.generate(stylesheet) : text;
}
// TODO: change this to a Duplex stream; remove the `istream` parameter.
//
// Defer this change until we drop support for Node.js v14. (This will be when
// moves to a more recent version of Electron, that upgrades to Node.js v16.)
// The implementation will become simpler at that point.
function makeCssRewriterStream(istream, base, encoding) {
    if (encoding === 'gzip') {
        const z = zlib_1.default.createGzip();
        const uz = zlib_1.default.createGunzip();
        const ws = makeCssRewriterStream(uz, base);
        istream.pipe(uz);
        ws.pipe(z);
        return z;
    }
    async function* iter() {
        const text = await fromReadable(istream);
        yield replaceUrlsInCss(text.toString());
    }
    return stream_1.Readable.from(iter());
}
//#endregion
//#region stream helpers
/** Read the remaining chunks from a ReadableStream, and combine them into a
 * single string (if they are all strings) or Buffer.
 *
 * Note: Doesn't handle chunks of type `Uint8Array`.
 */
async function fromReadable(stream, emptyValue = '') {
    const chunks = [];
    for await (const chunk of stream) {
        assert.ok(typeof chunk === 'string' || Buffer.isBuffer(chunk));
        chunks.push(chunk);
    }
    return chunks.length === 0 ? emptyValue
        : chunks.length === 1 ? chunks[0]
            : chunks.every(chunk => typeof chunk === 'string') ? chunks.join('')
                : chunks.every(chunk => chunk instanceof Buffer) ? Buffer.concat(chunks)
                    : Buffer.concat(chunks.map(chunk => typeof chunk === 'string' ? Buffer.from(chunk) : chunk));
}
function multiplexStreamWriter(streams) {
    assert.notEqual(streams.length, 0);
    return new stream_1.default.PassThrough({
        write(chunk, encoding, callback) {
            let error = null;
            let count = streams.length;
            for (const stream of streams) {
                stream.write(chunk, encoding, (err) => {
                    error ?? (error = err); // invoke the callback with only the first error
                    if (--count === 0) {
                        callback(error);
                    }
                });
            }
        },
        final(callback) {
            let count = streams.length;
            for (const stream of streams) {
                stream.end(() => {
                    if (--count === 0) {
                        callback();
                    }
                });
            }
        }
    });
}
//#endregion
//#region helpers
/** Call `callback` for each URL in the CSS stylesheet. If `callback` returns a
 * value, replace the URL with that value. */
function cssForEachUrl(stylesheet, callback) {
    csstree.walk(typeof stylesheet === 'string' ? (0, css_tree_1.parse)(stylesheet) : stylesheet, {
        visit: 'Url',
        enter(node) {
            // csstree's node.value is a string, but the latest @types/css-tree (v1)
            // declares it as a node.
            //
            // eslint-disable-next-line @typescript-eslint/no-explicit-any
            const urlNode = node;
            const transformed = callback(urlNode.value);
            if (transformed) {
                urlNode.value = transformed;
            }
        }
    });
}
function dedupe(array) {
    return Array.from(new Set(array));
}
function isRelativeUrl(url) {
    return !/^[a-z]+:/i.test(url);
}
function removeHash(url) {
    return url.replace(/#.*/, '');
}
// Source: nodejs documentation for Url.resolve
function urlResolve(from, to) {
    const resolvedUrl = new URL(to, new URL(from, 'resolve://'));
    if (resolvedUrl.protocol === 'resolve:') {
        const { pathname, search, hash } = resolvedUrl;
        return pathname + search + hash;
    }
    return resolvedUrl.toString();
}
//#endregion
