"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
var _a;
Object.defineProperty(exports, "__esModule", { value: true });
exports.createSyntaxErrorJsReporter = exports.sourceViewTemplate = exports.markdownToHtmlPage = exports.markedOptions = exports.templateDir = void 0;
const fs_1 = __importDefault(require("fs"));
const core_1 = __importDefault(require("highlight.js/lib/core"));
const css_1 = __importDefault(require("highlight.js/lib/languages/css"));
const javascript_1 = __importDefault(require("highlight.js/lib/languages/javascript"));
const plaintext_1 = __importDefault(require("highlight.js/lib/languages/plaintext"));
const shell_1 = __importDefault(require("highlight.js/lib/languages/shell"));
const marked_1 = __importDefault(require("marked"));
const nunjucks_1 = __importDefault(require("nunjucks"));
const path_1 = __importDefault(require("path"));
const pug_1 = __importDefault(require("pug"));
const helpers_1 = require("../helpers");
exports.templateDir = path_1.default.join(__dirname, './templates');
core_1.default.registerLanguage('css', css_1.default);
core_1.default.registerLanguage('javascript', javascript_1.default);
core_1.default.registerLanguage('plaintext', plaintext_1.default);
core_1.default.registerLanguage('shell', shell_1.default);
const markdownPageTemplate = pug_1.default.compileFile(path_1.default.join(exports.templateDir, 'markdown.pug'));
exports.markedOptions = {
    smartypants: true,
    highlight(code, lang) {
        const language = core_1.default.getLanguage(lang) ? lang : 'javascript';
        return core_1.default.highlight(code, { language }).value;
    }
};
function markdownToHtmlPage(data) {
    const markdown = (0, marked_1.default)(data, exports.markedOptions);
    const title = data.match(/^#\s*(.+)\s*$/m)?.[1] ?? '';
    return markdownPageTemplate({ markdown, title });
}
exports.markdownToHtmlPage = markdownToHtmlPage;
exports.sourceViewTemplate = pug_1.default.compileFile(path_1.default.join(exports.templateDir, 'source-view.pug'));
class SyntaxErrorFormatter {
    render(filepath, error) {
        const { jsTemplateEnv, syntaxErrorTemplate, syntaxErrorJsTemplate } = SyntaxErrorFormatter;
        // eslint-disable-next-line @typescript-eslint/no-explicit-any
        const { line: lineNo, column } = error.loc;
        let message = error.message;
        const locationSuffix = ` (${lineNo}:${column})`;
        if (message.endsWith(locationSuffix)) {
            message = message.slice(0, -locationSuffix.length);
        }
        // extract, escape, and number the lines
        let lines = fs_1.default.readFileSync(filepath, 'utf-8').trimEnd().split('\n');
        const lineNumberColWidth = String(lines.length + 1).length + 2;
        lines = lines.map((line, index) => {
            let lineLabel = String(1 + index);
            lineLabel = ' '.repeat(lineNumberColWidth - lineLabel.length) + lineLabel;
            if (1 + index === lineNo) {
                lineLabel = '<span style="color:red">▶︎</span> ' + lineLabel.slice(2);
            }
            return `${lineLabel} │ ${(0, helpers_1.escapeHTML)(line)}`;
        });
        // add a line beneath the error line
        lines.splice(lineNo, 0, ' '.repeat(lineNumberColWidth + 1) +
            '│' +
            ' '.repeat(column + 1) +
            `<span style="color:red">▲</span>`);
        const contextWindow = 8;
        // window the lines to the context around the error line
        lines.splice(lineNo + contextWindow);
        lines.splice(0, lineNo - contextWindow);
        // render HTML
        const contextHtml = syntaxErrorTemplate({
            error: { ...error, message },
            filepath,
            fullpath: path_1.default.resolve(filepath),
            context: lines.join('\n')
        });
        // return JS that will set the page body to the HTML
        return jsTemplateEnv.renderString(syntaxErrorJsTemplate, {
            fileName: path_1.default.basename(filepath),
            error,
            contextHtml
        });
    }
}
_a = SyntaxErrorFormatter;
SyntaxErrorFormatter.jsTemplateEnv = new nunjucks_1.default.Environment(null, {
    autoescape: false
});
SyntaxErrorFormatter.syntaxErrorTemplate = pug_1.default.compileFile(path_1.default.join(exports.templateDir, 'syntax-error.pug'));
SyntaxErrorFormatter.syntaxErrorJsTemplate = fs_1.default.readFileSync(path_1.default.join(exports.templateDir, 'report-syntax-error.js.njk'), 'utf-8');
(() => {
    _a.jsTemplateEnv.addFilter('quote', JSON.stringify);
})();
function createSyntaxErrorJsReporter(filepath, [error]) {
    return new SyntaxErrorFormatter().render(filepath, error);
}
exports.createSyntaxErrorJsReporter = createSyntaxErrorJsReporter;
